__all__ = ['PsuCtlError', 'run_action', 'Endpoint']

from typing import Callable, Dict, Optional, Tuple, Union

from edge import eventlog
from edge.support import fs
from edge.webserver.sysconf.uiconf.psuctl import PsuCtlUiconf

from psuctl._firmware import Firmware, VitekFirmware, StubFirmware
from psuctl._led_unit import factory as led_unit_factory
from psuctl._psu_unit import factory as psu_unit_factory
from psuctl._stepper_unit import factory as stepper_unit_factory
from psuctl._stepper_vitek_unit import factory as stepper_vitek_unit_factory
from psuctl._sh_unit import sh_unit_factory, sh_ics_unit_factory
from psuctl.exceptions import PsuCtlError
from psuctl.types import IFirmware, IFirmware2, IUnit, UnitType, UnitRef


_CONF_PATH = '/etc/edge/psuctl.conf'

UnitFactory = Callable[[Union[IFirmware, IFirmware2], UnitRef], Optional[IUnit]]

_UNIT_TYPE_TO_FACTORY_MAP: Dict[UnitType, UnitFactory] = {
    'led': led_unit_factory,
    'psu': psu_unit_factory,
    'stepper': stepper_unit_factory,
    'stepper-vitek': stepper_vitek_unit_factory,
    'sh': sh_unit_factory,
    'sh-ics': sh_ics_unit_factory,
}

_UNIT_TYPE_TO_FIRMWARE_MAP = {
    'led': Firmware,
    'psu': Firmware,
    'stepper': Firmware,
    'stepper-vitek': VitekFirmware,
    'sh': StubFirmware,
    'sh-ics': StubFirmware
}


@fs.Lock('/var/run/edge/psuctl.pid', try_count=3)
def run_action(script_name: str, action_conf: Dict) -> None:
    jnode = fs.load_json3(_CONF_PATH, default=None)
    if jnode is None:
        eventlog.log_e('fail: load conf')
        return

    scripts_jnode: Dict = jnode.get('scripts') or dict()
    script_jnode = scripts_jnode.get(script_name)
    if script_jnode is None:
        eventlog.log_e_param('#/scripts/{}'.format(script_name))
        return

    unit_type_ref = script_jnode.get('unit')
    if type(unit_type_ref) not in [list]:
        eventlog.log_e_param('#/scripts/{}/unit'.format(script_name))
        return

    elif not len(unit_type_ref):
        eventlog.log_e_param('#/scripts/{}/unit'.format(script_name))
        return

    unit_type = unit_type_ref[0]
    unit_factory = _UNIT_TYPE_TO_FACTORY_MAP.get(unit_type)
    if unit_factory is None:
        eventlog.log_e_param('#/scripts/{}/unit'.format(script_name))
        return

    endpoint = PsuCtlUiconf.get_endpoint()
    try:
        firmware_factory = _UNIT_TYPE_TO_FIRMWARE_MAP.get(unit_type)
        firmware = firmware_factory(endpoint)
    except PsuCtlError:
        eventlog.log_e('fail: firmware_factory')
        return

    unit = unit_factory(firmware, unit_type_ref)
    if unit is None:
        eventlog.log_e('fail: unit_factory')
        return

    unit_conf = script_jnode.get('conf') or dict()
    unit_conf.update(action_conf)
    unit.run(unit_conf)


class Endpoint:
    def __init__(self):
        jnode = fs.load_json3(_CONF_PATH, default=None)
        if jnode is None:
            eventlog.log_e('fail: load conf')
            raise PsuCtlError

        self._scripts_jnode: Dict = jnode.get('scripts') or dict()
        self._firmware = None
        self._firmware_factory = None

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    @fs.Lock('/var/run/edge/psuctl.pid', try_count=3)
    def run_action(self, script_name: str, action_conf: Dict) -> Optional[Tuple[int, int]]:
        script_jnode = self._scripts_jnode.get(script_name)
        if script_jnode is None:
            eventlog.log_e_param('#/scripts/{}'.format(script_name))
            raise PsuCtlError

        unit_type_ref = script_jnode.get('unit')
        if type(unit_type_ref) not in [list]:
            eventlog.log_e_param('#/scripts/{}/unit'.format(script_name))
            raise PsuCtlError

        elif not len(unit_type_ref):
            eventlog.log_e_param('#/scripts/{}/unit'.format(script_name))
            raise PsuCtlError

        unit_type = unit_type_ref[0]

        if self._firmware is None:
            endpoint = PsuCtlUiconf.get_endpoint()

            try:
                self._firmware_factory = _UNIT_TYPE_TO_FIRMWARE_MAP.get(unit_type)
                self._firmware = self._firmware_factory(endpoint)
            except PsuCtlError:
                eventlog.log_e('fail: firmware_factory')
                return

        if not isinstance(self._firmware, self._firmware_factory):
            eventlog.log_e('fail: unit {}'.format(unit_type))
            raise PsuCtlError

        unit_factory = _UNIT_TYPE_TO_FACTORY_MAP.get(unit_type)
        if unit_factory is None:
            eventlog.log_e_param('#/scripts/{}/unit'.format(script_name))
            raise PsuCtlError

        unit = unit_factory(self._firmware, unit_type_ref)
        if unit is None:
            eventlog.log_e('fail: unit_factory')
            raise PsuCtlError

        unit_conf = script_jnode.get('conf') or dict()
        unit_conf_copy = unit_conf.copy()
        unit_conf_copy.update(action_conf)
        return unit.run(unit_conf_copy)
