import serial
import time

from typing import Dict, Optional, Union

from pymodbus.client.serial import ModbusSerialClient
from pymodbus.exceptions import ModbusException
from pymodbus.pdu import ModbusExceptions, ExceptionResponse
from pymodbus.register_read_message import ReadInputRegistersResponse
from pymodbus.register_write_message import WriteSingleRegisterResponse
from pymodbus.bit_write_message import WriteSingleCoilResponse
from pymodbus.bit_read_message import ReadCoilsResponse

from edge import eventlog
from psuctl.exceptions import PsuCtlError
from psuctl.types import IFirmware, IFirmware2


class Firmware(IFirmware):
    _SLAVE_ADDRESS = 0x30

    _RX_TO_TX_TOUT_SEC = 0.025

    def __init__(self, endpoint: Optional[str] = None):
        if endpoint is None:
            eventlog.log_e('fail: endpoint is not provided')
            raise PsuCtlError

        IFirmware.__init__(self)
        self._modbus = ModbusSerialClient(
            method="rtu", port=endpoint, baudrate=19200,
            bytesize=8, stopbits=1, parity='N')
        self._version = None

    def read_ro_reg(self, address: int, ack_tout_sec: float) -> Optional[int]:
        """
        :raise PsuCtlError:
        """
        self._modbus.timeout = ack_tout_sec
        try:
            with self._modbus as modbus:
                reply = modbus.read_input_registers(address, count=1, unit=Firmware._SLAVE_ADDRESS)

            time.sleep(self._RX_TO_TX_TOUT_SEC)
            if reply is None:
                eventlog.log_d('read_input_registers <{}> - no such register)'.format(address))
                return None

            elif isinstance(reply, ReadInputRegistersResponse):
                reg0 = reply.registers[0]

                # Fix byte-order. Firmwares v1.1, v1.2, v1.3 and v1.4 responds
                # in little-endian, instead of big-endian.
                value = ((reg0 >> 8) & 0xFF) | ((reg0 << 8) & 0xFF00)

            elif isinstance(reply, ReadCoilsResponse):
                # Firmware v1.1 responds with a ReadCoilsResponse for a ReadInputRegistersRequest
                #
                # Fix byte-order. Firmwares v1.1, v1.2, v1.3 and v1.4 responds
                # in little-endian, instead of big-endian.
                value = int(''.join(reversed(['1' if bit else '0' for bit in reply.bits])), 2)

            elif not isinstance(reply, ExceptionResponse):
                eventlog.log_e('fail: read_input_registers (unsupported response <{}>)'.format(str(reply)))
                raise NotImplementedError

            elif reply.exception_code == ModbusExceptions.IllegalAddress:
                eventlog.log_w('fail: read_input_registers (unsupported register)')
                return None

            else:
                eventlog.log_e('fail: read_input_registers ({})'.format(str(reply)))
                raise PsuCtlError

            eventlog.log_d('done: read_input_registers (addres:{}, value:{})'.format(address, value))
            return value

        except ModbusException as x:
            eventlog.log_e('fail: read_input_registers ({})'.format(str(x)))
            raise PsuCtlError

    def write_rw_reg(self, address: int, int16_value: int, ack_tout_sec: float) -> int:
        """
        :returns:
            int16_value
        :raise PsuCtlError:
        """
        eventlog.log_d('exec: write_register (address: {}, value{})'
                       .format(hex(address), hex(int16_value)))

        self._modbus.timeout = ack_tout_sec
        try:
            with self._modbus as modbus:
                reply = modbus.write_register(address, int16_value, unit=Firmware._SLAVE_ADDRESS)

            time.sleep(Firmware._RX_TO_TX_TOUT_SEC)
            if reply is None:
                eventlog.log_w('fail: write_register timeout (tout-sec: {})'.format(ack_tout_sec))
                raise PsuCtlError

            elif not isinstance(reply, WriteSingleRegisterResponse):
                eventlog.log_d('done: write_register (invalid response)')
                raise PsuCtlError

            elif reply.value != int16_value:
                eventlog.log_w('fail: write_register (address:{}, query-value:{}, reply-value:{})'
                               .format(hex(address), hex(int16_value), hex(reply.value)))

            else:
                eventlog.log_d('done: write_register (address:{}, value:{})'
                               .format(hex(address), hex(int16_value)))

            return reply.value

        except ModbusException as x:
            eventlog.log_e('fail: write_register ({})'.format(str(x)))
            raise PsuCtlError

    def write_rw_bit(self, address: int, int1_value: Union[bool, int], ack_tout_sec: float) -> int:
        """
        :returns:
            int1_value

        :raise PsuCtlError:
        """
        self._modbus.timeout = ack_tout_sec
        try:
            with self._modbus as modbus:
                reply = modbus.write_coil(address, int1_value, unit=Firmware._SLAVE_ADDRESS)

            time.sleep(Firmware._RX_TO_TX_TOUT_SEC)
            if reply is None:
                eventlog.log_w('fail: write_coil timeout (tout-sec: {})'.format(ack_tout_sec))
                raise PsuCtlError

            elif not isinstance(reply, WriteSingleCoilResponse):
                eventlog.log_d('done: write_coil (invalid response)')
                raise PsuCtlError

            elif reply.value != int1_value:
                eventlog.log_w('fail: write_coil (address:{}, query-value:{}, reply-value:{})'
                               .format(hex(address), hex(int1_value), hex(reply.value)))

            else:
                eventlog.log_d('done: write_coil (address:{}, value:{})'
                               .format(hex(address), hex(int1_value)))

            return reply.value

        except ModbusException as x:
            eventlog.log_e('fail: write_coil ({})'.format(str(x)))
            raise PsuCtlError

    @property
    def version(self) -> str:
        if self._version is None:
            try:
                value = self.read_ro_reg(address=0xFFFF, ack_tout_sec=0.25) or 0x10

            except PsuCtlError:
                eventlog.log_e('fail: read_ro_reg')
                raise

            # map 0x14 to 'v1.4'
            self._version = 'v{}.{}'.format((value // 16) % 16, value % 16)

        return self._version


class VitekFirmware(IFirmware2):
    REF_CHAN1 = 'ch1'
    REF_CHAN2 = 'ch2'
    REF_CHAN3 = 'ch3'

    def __init__(self, endpoint: Optional[str] = None):
        if endpoint is None:
            eventlog.log_e('fail: endpoint is not provided')
            raise PsuCtlError

        IFirmware2.__init__(self)
        # if pass endpoint to constructor it will try to open port and may throw exeption
        self._port = serial.Serial(port=endpoint, baudrate=9600, parity=serial.PARITY_NONE, stopbits=1)
        # self.ser.port(endpoint)
        # self._port.close()

        self._caps: Optional[Dict[str, Dict[str, int]]] = None
        self._stat: Optional[Dict[str, Dict[str, int]]] = None

    def write(self, query_msg: str) -> int:
        try:
            self._port.reset_input_buffer()
            return self._port.write(query_msg)

        finally:
            self._stat = None

    def read(self) -> str:
        reply_msg = self._port.read_all()
        return reply_msg

    def _refresh(self) -> None:
        self._port.write('?')
        time.sleep(1.5)
        reply = self._port.read_all()

        def get(param: str) -> Optional[int]:
            param_pos = reply.find(param)
            if -1 == param_pos:
                eventlog.log_e('fail: find start (param:{})'.format(param))
                raise PsuCtlError

            param_pos += len(param)
            endlpos = reply.find('\n', param_pos)
            if -1 == endlpos:
                eventlog.log_e('fail: find end (param:{})'.format(param))
                raise PsuCtlError

            # skip ":" after param-name
            param_value = reply[param_pos+1:endlpos]
            return int(param_value)

        ch1_now = get('Ch1 steps now')
        ch1_max = get('Full range Ch1 steps')
        ch2_now = get('Ch2 steps now')
        ch2_max = get('Full range Ch2 steps')
        ch3_now = get('Ch3 steps now')
        ch3_max = get('Full range Ch3 steps')
        step_tout1 = get('Stepper motor Ch1 period, x100us')
        step_tout2 = get('Stepper motor Ch2 period, x100us')
        step_tout3 = get('Stepper motor Ch3 period, x100us')
        if None in [ch1_now, ch1_max, ch2_now, ch3_max, ch3_now, ch2_max, step_tout1, step_tout2, step_tout3]:
            raise PsuCtlError

        self._caps = {
            self.REF_CHAN1: {
                self.CAP_CHAN_MAX_STEPS: ch1_max,
                self.CAP_CHAN_STEP_TOUT: step_tout1
            },
            self.REF_CHAN2: {
                self.CAP_CHAN_MAX_STEPS: ch2_max,
                self.CAP_CHAN_STEP_TOUT: step_tout2
            },
            self.REF_CHAN3: {
                self.CAP_CHAN_MAX_STEPS: ch3_max,
                self.CAP_CHAN_STEP_TOUT: step_tout3
            }
        }

        self._stat = {
            self.REF_CHAN1: {
                self.STAT_CHAN_STEPS: ch1_now
            },
            self.REF_CHAN2: {
                self.STAT_CHAN_STEPS: ch2_now
            },
            self.REF_CHAN3: {
                self.STAT_CHAN_STEPS: ch3_now
            }
        }

    def get_chan_cap(self, chan_ref, cap_name) -> int:
        if self._caps is None:
            self._refresh()

        return self._caps[chan_ref][cap_name]

    def get_chan_stat(self, chan_ref, stat_name) -> int:
        if self._stat is None:
            self._refresh()

        return self._stat[chan_ref][stat_name]


class StubFirmware(IFirmware):
    def __init__(self, endpoint=None):
        IFirmware.__init__(self)

    def write_rw_reg(self, address: int, int16_value: int, ack_tout_sec: float) -> int:
        """
        :returns:
            int16_value
        :raise PsuCtlError:
        """
        raise PsuCtlError

    def write_rw_bit(self, address: int, int1_value: Union[bool, int], ack_tout_sec: float) -> int:
        """
        :returns:
            int1_value

        :raise PsuCtlError:
        """
        raise PsuCtlError

    @property
    def version(self) -> str:
        return 'self._version'
