__all__ = ['sh_unit_factory', 'sh_ics_unit_factory']

from typing import Dict, Optional, Tuple, Union

from edge import eventlog
from edge.support import sh

from psuctl.types import IFirmware, IFirmware2, IUnit, UnitRef
from psuctl.exceptions import PsuCtlError


class _ShUnit(IUnit):
    def __init__(self, firmware: Optional[Union[IFirmware, IFirmware2]], ref: UnitRef):
        super().__init__(firmware, ref)

    def run(self, conf: Dict) -> Optional[Tuple[int, int]]:
        """
        :param conf:
            psuctl.conf#/scripts/<name>/conf
        """
        shell_cmd = conf.get('shell')
        if shell_cmd and type(shell_cmd) not in [str]:
            eventlog.log_e_param('#/shell')
            raise PsuCtlError

        try:
            rc = sh.run_silent(shell_cmd)
            if rc == 0:
                eventlog.log_d('done: {}'.format(shell_cmd))
            else:
                eventlog.log_d('fail: {} rc {}'.format(shell_cmd, rc))

        except Exception:
            eventlog.log_x()
            raise PsuCtlError

        return None


class _ShIcsUnit(IUnit):
    def __init__(self, firmware: Optional[Union[IFirmware, IFirmware2]], ref: UnitRef):
        super().__init__(firmware, ref)

    def run(self, conf: Dict) -> Optional[Tuple[int, int]]:
        """
        :param conf:
            psuctl.conf#/scripts/<name>/conf
        """
        action = conf.get('action')

        if action not in ['rotate', 'get-limits']:
            eventlog.log_e_param('#/action')
            raise PsuCtlError

        if action == 'get-limits':
            ch_max = conf.get('max-steps')
            ch_min = conf.get('min-steps', 0)
            if ch_max is None:
                eventlog.log_e('fail:#/max-steps')
                raise PsuCtlError
            return ch_min, ch_max

        raw_steps = conf.get('steps')
        if type(raw_steps) in [str]:
            raw_steps = int(raw_steps)
        elif type(raw_steps) in [int]:
            pass
        else:
            eventlog.log_e_param('#/steps')
            raise PsuCtlError

        # rot_steps = abs(raw_steps)

        shell_cmd = conf.get('shell')
        if shell_cmd and type(shell_cmd) not in [str]:
            eventlog.log_e_param('#/shell')
            raise PsuCtlError

        shell_cmd += ' {}'.format(raw_steps)
        try:
            rc = sh.run_silent(shell_cmd)
            if rc == 0:
                eventlog.log_d('done: {}'.format(shell_cmd))
            else:
                eventlog.log_d('fail: {} rc {}'.format(shell_cmd, rc))
                raise PsuCtlError

        except Exception:
            eventlog.log_x()
            raise PsuCtlError


def sh_unit_factory(
        firmware: Optional[Union[IFirmware, IFirmware2]],
        unit_ref: Optional[Tuple[str, str]] = None
        ) -> Optional[IUnit]:

    return _ShUnit(firmware, unit_ref)


def sh_ics_unit_factory(
        firmware: Optional[Union[IFirmware, IFirmware2]],
        unit_ref: Optional[Tuple[str, str]] = None
        ) -> Optional[IUnit]:

    return _ShIcsUnit(firmware, unit_ref)
